open VecOps

(*

Polygon-fill algorithm

0. We have a list of coordinates, ordered in counter-clockwise fashion, which are converted 
   to segments, connecting the last coordinate to the first in the process.

1. We create left and right list. 

   The left list has segments that go from up to down, the right list has the segments that
   go from down to up. Horizontal segments are removed.
     orig          left      right
         /\            /       \
     /\_/  \  -->  /  /    \    \
     \     /       \            /
      \___/         \          /

2. Segments in the right list (segments from down to up) are reversed to go from up to down

3. Segments are converted into slopes and they are sorted with primary key being current y and 
   secondary key being current x

4. If at this point we have two non-empty lists, continue, else end algorithm

5. We pick the first elements from the both of the lists

6. Both picked elements have their current y-location same (assertion)

7. Fill rows between y..y2'min

8. Remove rows for which their current row is the same as their y2

9. Continue from step 3

*)

let cmp_seg_horiz ((x1, _), _) ((x2, _), _) =
  x1 - x2;;

let rec connect_hd_tl list =
  let a = List.hd list in
  list @ [ a ];;

type direction = DOWN | UP;;

let direction segment = 
  match segment with
    ((_, y1), (_, y2)) ->
      if y1 < y2 then DOWN else UP;;

let swap_direction (a, b) = (b, a);;

let find_next_y segs =
  let rec find segs curmin =
    match segs with
      [] -> ( 
	match curmin with
	  None -> failwith ":(" (* raise Not_found *)
	| Some curmin -> curmin
       )
    | (_, (_, y))::tl ->
	match curmin with
	  None -> 
	    find tl (Some y)
	| Some min ->
	    let _curmin = if y < min then y else min in
	    find tl (Some min)
  in
  find segs None;;

let output_seg seg = 
  fun chan ->
    match seg with 
      ((x1, y1), (x2, y2)) -> Printf.fprintf chan "(%d, %d)->(%d, %d)" x1 y1 x2 y2;;

type slope = Slope of (unit -> ((int * int) * slope));;

type work = { slope: slope; at: (int * int); end_y: int };;

let debug = ref false;;

let slope ((x1, y1), (x2, y2)) =
  assert (y1 <= y2);
  let dx = x2 - x1 in
  let dy = y2 - y1 in
  let fsx = (float_of_int dx) /. (float_of_int dy) in

  let rec draw n =
    if n <= (y2 - y1) then (
      let out_x = (int_of_float ((float_of_int n) *. fsx)) + x1 in
      let out_y = y1 + n in
      if !debug then Printf.printf "Slope %t at (%d, %d)\n%!" (output_seg ((x1, y1), (x2, y2))) out_x out_y else ();
      ((out_x, out_y), Slope (fun () -> draw (n + 1)))
     )
    else
      raise Not_found
  in
  draw 0
;;

let slope_next slope =
  match slope with
    Slope slope -> slope ();;

let begin_work segment =
  match slope segment with
    (at, next) ->
      match segment with (_, (_, end_y)) ->
	{ slope = next; at = at; end_y = end_y };;

let do_work work =
  match work.slope with
    Slope next ->
      match next () with
	(at, next) ->
	  { work with slope = next; at = at; };;


let print_filler y x1 x2 =
  Printf.printf "row %d %d - %d\n%!" y x1 x2;;

let fill_work left right y'min filler =
  assert (snd(left.at) = snd(right.at));
  let rec work left right max_y =
    let () = filler (snd left.at) (fst left.at) (fst right.at) in
      if (snd(left.at)) < max_y then
	let _ = () in
	  (* Printf.printf "Filling segment between (%d,%d) and (%d,%d)\n%!" (fst left.at) (snd left.at) (fst right.at) (snd right.at); *)
	  work (do_work left) (do_work right) max_y
      else
	(left, right)
  in
    work left right y'min;;
    
let compare_work a b =
  match a with { at = (ax, ay) } ->
    match b with { at = (bx, by) } ->
      let c = ay - by in
      if c <> 0 then
	c
      else
	ax - bx;;

let work_valid work =
  match work with
    { at = (_, y); end_y = end_y } ->
      y < end_y;;

let output_work work output =
  Printf.fprintf output "at (%d, %d), end_y = %d" (fst work.at) (snd work.at) work.end_y;;

let next_y segments limit_y min_y =
  List.fold_left 
    ( fun y work ->
      let work_y = (snd work.at) in
      if (work_y > limit_y) & (work_y < y) then
	work_y
      else
	y
     ) min_y segments;;

let rec scan (left, right) filler =
  let left = List.sort compare_work left in
  let right = List.sort compare_work right in
  if !debug then List.iter (fun x -> Printf.printf "left: %t\n%!" (output_work x)) left else ();
  if !debug then List.iter (fun x -> Printf.printf "right: %t\n%!" (output_work x)) right else ();

  if ((left <> []) || (right <> [])) then
    match (left, right) with
      (l :: ltl, r :: rtl) when snd(l.at) = snd(r.at) ->
	let cur_y = snd l.at in
	let y'min = min l.end_y r.end_y in
	let y'min = min (next_y ltl cur_y y'min) (next_y rtl cur_y y'min) in
	let (l, r) = fill_work l r y'min filler in
	((if work_valid l then l::ltl else ltl), 
	 (if work_valid r then r::rtl else rtl))
    | _ -> (left, right)
  else
    (left, right)
;;

let rec generate_segs vertices segments =
  match vertices with
    a :: b :: tail -> (
      match (a, b) with
	((_, y1), (_, y2)) when y1 = y2 -> 
	  generate_segs (b::tail) segments
      | _ -> 
	  generate_segs (b::tail) ((a, b)::segments)
     )
  | _ -> List.rev segments;;

let rec separate_segs segments (left, right) =
  match segments with
    seg :: tail -> 
      let dir =  direction seg in
      let seg = match dir with UP -> swap_direction seg | DOWN -> seg in
      let list = match dir with DOWN -> left | UP -> right in
      let list = seg :: list in
      separate_segs tail (match dir with DOWN -> (list, right) | UP -> (left, list))
  | _ -> (left, right);;


let output_segs segs output =
  List.iter 
    ( fun ((a, b), (c, d)) -> Printf.fprintf output "(%d, %d) -> (%d, %d)\n" a b c d )
    segs;;

let rec polygon_work (left, right) filler =
  if (left <> []) & (right <> []) then
    let (left, right) = scan (left, right) filler in
    polygon_work (left, right) filler
  else
    ();;

let polygon vertices filler =
  if List.length vertices < 3 then
    ()
  else
    let vertices = connect_hd_tl vertices in
    let segs = generate_segs vertices [] in
    let (left, right) = separate_segs segs ([], []) in
      if !debug then Printf.printf "Left\n%t%!" (output_segs left) else ();
      if !debug then Printf.printf "Right\n%t%!" (output_segs right) else ();

      polygon_work
	((List.sort compare_work (List.map begin_work left)),
	 (List.sort compare_work (List.map begin_work right)))
	filler
;;

let within ((x1, y1), (x2, y2)) (x, y) =
  x >= x1 && y >= y1 && x < x2 && y < y2

let not1 f a = 
  not (f a)

let sign = function
  | n when n < 0 -> -1
  | n -> 1

(* lazy, got the formulas from http://local.wasp.uwa.edu.au/~pbourke/geometry/lineline2d/ *)
let intersect_u ((x1, y1), (x2, y2)) ((x3, y3), (x4, y4)) =
  let dem = ((y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1)) in
    if dem = 0 then
      None
    else
      let u'a = float ((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3)) /. float dem in
      let u'b = float ((x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3)) /. float dem in
	Some ((u'a, lazy ((x1 + int_of_float (u'a *. float (x2 - x1))),
			  (y1 + int_of_float (u'a *. float (y2 - y1))))), 
	      (u'b, lazy ((x3 + int_of_float (u'a *. float (x4 - x3))),
			  (y3 + int_of_float (u'a *. float (y4 - y3))))))

let intersect_line a b =
  match intersect_u a b with
    | None -> None
    | Some ((u'a, p), (u'b, _)) 
	when u'a >= 0.0 && u'a <= 1.0 && u'b >= 0.0 && u'b <= 1.0 ->
	Some (Lazy.force p)
    | _ -> None

(* the line a is considered of being infinitely long *)
let intersect_line_any a b =
  match intersect_u a b with
    | None -> None
    | Some ((u'a, p), (u'b, _)) 
	when u'b >= 0.0 && u'b <= 1.0 ->
	Some (Lazy.force p)
    | _ -> None

(* lazy, got the formulas from http://local.wasp.uwa.edu.au/~pbourke/geometry/lineline2d/ *)
let intersect_line_float ((x1, y1), (x2, y2)) ((x3, y3), (x4, y4)) =
  let dem = ((y4 -. y3) *. (x2 -. x1) -. (x4 -. x3) *. (y2 -. y1)) in
    if dem = 0.0 then
      None
    else
      let u'a = ((x4 -. x3) *. (y1 -. y3) -. (y4 -. y3) *. (x1 -. x3)) /. dem in
      let u'b = ((x2 -. x1) *. (y1 -. y3) -. (y2 -. y1) *. (x1 -. x3)) /. dem in
	if u'a >= 0.0 && u'a <= 1.0 && u'b >= 0.0 && u'b <= 1.0 then
	  Some ((x1 +. u'a *. (x2 -. x1)),
		(y1 +. u'a *. (y2 -. y1)))
	else
	  None

(* returns intersecting points in the order of intersection *)
let intersect_region_with intersector ((x1, y1), (x2, y2)) ((p, _) as line) =
  let d2 (x1, y1) = 
    let (x2, y2) = p in
      (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1)
  in
  List.sort 
    (fun p1 p2 -> compare (d2 p1) (d2 p2))
    (List.concat (
       (List.map (fun border -> 
		    match intersector border line with
		      | None -> []
		      | Some x -> [x])
	  [(x1, y1), (x1, y2 - 1);
	   (x1, y2 - 1), (x2 - 1, y2 - 1); 
	   (x2 - 1, y2 - 1), (x2 - 1, y1);
	   (x2 - 1, y1), (x1, y1)])
     )
    )


let intersect_region = intersect_region_with intersect_line

let intersect_borders = intersect_region_with intersect_line_any


let rec init l =
  match l with
    | a::((_::_) as rest) ->
	a::init rest
    | a::[] ->
	[]
    | [] -> failwith "Empty list"

let rec last l =
  match l with
    | _::(last::[]) ->
	last
    | _::rest ->
	last rest
    | [] -> failwith "Empty list"

let debug = ref false

let str_vs vs =
  (String.concat ";" (List.map Vector.i_to_string vs))

let str_segs segs =
  (String.concat ";" (List.map (fun (a, b) -> Vector.i_to_string a ^ "-" ^ Vector.i_to_string b) segs))

let rec segments_of_vertices vs =
  match vs with
    | a::((b::_) as rest) ->
	(a, b)::segments_of_vertices rest
    | _ -> []

(* still misses the case that can leave a corner unfilled (X in the image):

         /\
   +----/--\
   |   /  X|\
   |  /____|_\
   |       |
   +-------+

          
   +----/--+
   |   /  \| 
   |  /____\  
   |       |
   +-------+ *)

let clip_rect vertices (((x1, y1), (x2, y2)) as region) =
  let within = within region in
    if !debug then
      Printf.printf "Clipping vertices [%s] with region ((%d,%d),(%d,%d))\n"
	(String.concat ";" (List.map Vector.i_to_string vertices))
	x1 y1 x2 y2;
    match vertices with
      | [] -> []
	  (*      | vertices when List.for_all (not1 within) vertices ->
		  [(x1, y1); (x1, y2); (x2, y2); (x2, y1)]*)
      | vertices ->
	  let vertices = connect_hd_tl vertices in
	  let segs = segments_of_vertices vertices in
	  let segs =
	    List.concat (
	      List.map 
		(fun (a, b) ->
		   match within a, within b with
		     | true, true -> 
			 if !debug then Printf.printf "1\n";
			 [(a, b)]
		     | false, false -> 
			 if !debug then Printf.printf "2\n";
			 (match intersect_region region (a, b) with
			    | [] | _::[] -> []
			    | vs -> (* must be more than one intersection *)
				segments_of_vertices vs)
		     | true, false -> 
			 if !debug then Printf.printf "3\n";
			 let vs = a::intersect_region region (a, b) in
			 segments_of_vertices vs
		     | false, true ->
			 let vs = intersect_region region (a, b) @ [b] in
			 if !debug then Printf.printf "4 : %s, %s: %s -> %s\n" (Vector.i_to_string a)  (Vector.i_to_string b) (str_vs vs) (str_segs (segments_of_vertices vs));
			   segments_of_vertices vs
		)
		segs
	    )

	  in

	  let vs = 
	    List.concat (
	      List.map
		(fun (a, b) -> [a; b])
		segs
	    )
	  in
	  let vs = 
	    let rec filter_dups vs =
	      match vs with
		| a::((b::_) as rest) when a = b ->
		    filter_dups rest
		| a::rest ->
		    a::filter_dups rest
		| [] -> []
	    in
	      filter_dups vs
	  in
	    if !debug then Printf.printf "%!";
	    (*segs', segs, *)
	  vs

